#ifndef SLICEDISTORTTRANSFORMER_H
#define SLICEDISTORTTRANSFORMER_H

#include <cmath>

#include "enum.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicenormalizetransformer.h"
#include "sliceffttransformer.h"


template<class T>
class SliceDistortTransformer : public SliceTransformer<T>
{
public:
    SliceDistortTransformer (
        double cutoff,
        double distortRange,
        double amplificationFactor,
        int distortType = DISTORT_PROPORTIONAL,//DISTORT_70SFUZZGUITAR
        size_t feedback = 0,
        double toneShaping1 = 1.0,
        double toneShaping2 = 0.0,
        size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T> *channel );
private:
    double cutoff, distortRange, amplificationFactor;
    int distortType;
    size_t feedback;
    double toneShaping1, toneShaping2;
    size_t channels;
    //double ( *nonlinearity ) ( double decibel );
    T  distortSampleProportional( T sample, size_t pos );
};

template<class T>
SliceDistortTransformer<T>::SliceDistortTransformer (
    double cutoffParameter,
    double distortRangeParameter,
    double amplificationFactorParameter,
    int distortTypeParameter,
    size_t feedbackParameter,
    double toneShapingParameter1,
    double toneShapingParameter2,
    size_t channelsParameter ) :
        cutoff ( cutoffParameter ),
        distortRange ( distortRangeParameter ),
        amplificationFactor ( amplificationFactorParameter ),
        distortType ( distortTypeParameter ),
        feedback ( feedbackParameter ),
        toneShaping1 ( toneShapingParameter1 ),
        toneShaping2 ( toneShapingParameter2 ),
        channels ( channelsParameter )
{

    if ( distortRange > cutoff)
        throw;
    if ( amplificationFactor < 0 )
        throw;
    if ( ( amplificationFactor * cutoff ) > SHRT_MAX )
        throw;

    /*switch ( distortType )
    {
        case DISTORT_B3OVERDRIVE:
            nonlinearity = b3Overdrive;
            break;
        case DISTORT_FLATRESPONSE:
            nonlinearity = flatResponse;
            break;
        case DISTORT_70SFUZZGUITAR:
            nonlinearity = seventiesFuzzguitar;
            break;
        case DISTORT_TUBEDRIVE:
            nonlinearity = tubeDrive;
            break;
        default:
            nonlinearity = tubeDrive;
            break;
    }*/
}


template<class T>
void SliceDistortTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    for ( size_t i = 0; i < feedback; i++ )
    {
        SliceNormalizeTransformer<T> snt;
        snt.transform ( leftChannel, rightChannel );

        SliceFFTTransformer<T> lpf (
            FFT_LOWPASS,
            0.0,
            0.0,
            toneShaping1,
            toneShaping2
        );
        lpf.transform ( leftChannel, rightChannel );

        transformChannel ( leftChannel );
        if ( rightChannel )
            transformChannel ( rightChannel );
    }

    //tbd
    //SliceFFTTransformer<T> hpf (
    //  FFT_HIGHPASS,
    //0.075,
    //0.075,
    //0.0,
    //0.0
    //);
    //hpf.transform ( leftChannel, rightChannel );
}

template<class T>
void SliceDistortTransformer<T>::transformChannel ( std::vector<T> *channel )
{
    size_t channelSize;
    channelSize = channel->size();

    //double decibelValue, newDecibelValue;

    for ( size_t i = 0; i < channelSize; i++ )
    {
        (*channel) [ i ] = ( distortSampleProportional ( channel->at ( i ), i ));
        /*decibelValue = SampleMath::amplitudeToDecibel ( channel->at ( i ) );
        newDecibelValue =  nonlinearity ( decibelValue );
        (*channel) [ i ] = SampleMath::decibelToAmplitude ( decibelValue );//newDecibelValue );*/
    }
}

template<class T>
T SliceDistortTransformer<T>::distortSampleProportional ( T sample, size_t pos )
{
    double absSample, linearRange;
    absSample =  fabs ( sample );
    linearRange = cutoff - distortRange;
    if ( ! ( absSample > linearRange ) )
        return amplificationFactor * sample;
    double distortFactor, absSampleDistorted, sinFactor;
    distortFactor = distortRange / ( SHRT_MAX - linearRange );
    sinFactor = sin ( 2 * PI / 2205 * pos );
    absSampleDistorted = linearRange + distortFactor * sinFactor * ( absSample - linearRange );
    return amplificationFactor * SampleMath::sign ( sample ) * absSampleDistorted;
}

#endif




// implementation using window technique

/*#ifndef SLICEDISTORTTRANSFORMER_H
#define SLICEDISTORTTRANSFORMER_H

#include <cmath>

#include "enum.h"
#include "samplemath.h"
#include "slicetransformer.h"
#include "slicenormalizetransformer.h"
#include "sliceffttransformer.h"


template<class T>
class SliceDistortTransformer : public SliceTransformer<T>
{
public:
    SliceDistortTransformer (
        double cutoff,
        double distortRange,
        double amplificationFactor,
        int distortType = DISTORT_PROPORTIONAL,
        size_t feedback = 0,
        double toneShaping1 = 1.0,
        double toneShaping2 = 0.0,
        size_t channels = 2 );
    void transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel );
    void transformChannel ( std::vector<T>& channel );
private:
    double cutoff, distortRange, amplificationFactor;
    int distortType;
    size_t feedback;
    double toneShaping1, toneShaping2;
    size_t channels;
    T  distortSampleProportional( T sample, size_t pos );
};

template<class T>
SliceDistortTransformer<T>::SliceDistortTransformer (
    double cutoffParameter,
    double distortRangeParameter,
    double amplificationFactorParameter,
    int distortTypeParameter,
    size_t feedbackParameter,
    double toneShapingParameter1,
    double toneShapingParameter2,
    size_t channelsParameter ) :
        cutoff ( cutoffParameter ),
        distortRange ( distortRangeParameter ),
        amplificationFactor ( amplificationFactorParameter ),
        distortType ( distortTypeParameter ),
        feedback ( feedbackParameter ),
        toneShaping1 ( toneShapingParameter1 ),
        toneShaping2 ( toneShapingParameter2 ),
        channels ( channelsParameter )
{

    if ( distortRange > cutoff)
        throw;
    if ( amplificationFactor < 0 )
        throw;
    if ( ( amplificationFactor * cutoff ) > SHRT_MAX )
        throw;
}


template<class T>
void SliceDistortTransformer<T>::transform ( std::vector<T> *leftChannel, std::vector<T> *rightChannel )
{
    for ( size_t i = 0; i < feedback; i++ )
    {
        SliceNormalizeTransformer<T> snt;
        snt.transform ( leftChannel, rightChannel );

        SliceFFTTransformer<T> hpf (
            FFT_LOWPASS,
            0.0,
            0.0,
            toneShaping1,
            toneShaping2
        );
        hpf.transform ( leftChannel, rightChannel );

        transformChannel ( *leftChannel );
        if ( rightChannel )
            transformChannel ( *rightChannel );
    }

    //tbd
    //SliceFFTTransformer<T> hpf (
      //  FFT_HIGHPASS,
        //0.075,
        //0.075,
        //0.0,
        //0.0
    //);
    //hpf.transform ( leftChannel, rightChannel );
}

template<class T>
void SliceDistortTransformer<T>::transformChannel ( std::vector<T>& channel )
{
    std::vector<T> result;

    typename std::vector<T>::iterator resultIt;

    size_t channelSize, windowSize, windowSizeHalf, crossFadeSize, windowOffset, windowNo;
    channelSize = channel.size();

    //std::cerr << "channelSize = " << channelSize << std::endl;

    windowSize = 1024;
    windowSizeHalf = windowSize / 2;
    crossFadeSize = windowSize / 2;
    windowOffset = windowSize - crossFadeSize;
    windowNo =  ( channelSize - windowSize ) / windowOffset + 1;

    double windowMean, decibelValue, decibelMin, decibelMax;

    for ( size_t j = 0; j < channelSize; j++ )
        result.push_back ( 0.0 );

    std::vector <T> currentWindow (windowSize);
    for ( size_t i = 0; i < windowNo; i++ )
    {
        currentWindow.clear();
        windowMean = 0;

        for ( size_t j = 0; j < windowSize; j++ )
        {
            windowMean += channel [ i * windowOffset + j ];
        }

        windowMean /= windowSize;
        decibelMin = 0.0;
        decibelMax = -52.0;
        for ( size_t j = 0; j < windowSize; j++ )
        {
            decibelValue = 20 * log10 ( channel [ i * windowOffset + j ] / windowMean );
            decibelMin = ( decibelValue < decibelMin ) ? decibelValue: decibelMin;
            decibelMax = ( decibelValue > decibelMax ) ? decibelValue: decibelMax;
            currentWindow.push_back ( distortSampleProportional ( channel [ i * windowOffset + j ], ( i * windowOffset ) + j ));
        }

        //std::cerr << "windowMean = " << windowMean << std::endl;
        //std::cerr << "decibelMin = " << decibelMin << std::endl;
        //std::cerr << "decibelMax = " << decibelMax << std::endl;

        for ( size_t j = 0; j < windowSize; j++ )
        {
            if ( i && ( j < crossFadeSize ) )
            {
                currentWindow [ j ] = lConcaveSoftClipper ( j, 0, crossFadeSize ) * currentWindow.at ( j );
            }
            else if ( ( i < ( windowNo - 1 ) ) && ( j > windowSize - crossFadeSize - 1 ) )
            {
                currentWindow [ j ] = rConvexSoftClipper ( j, windowSize - crossFadeSize, windowSize ) * currentWindow.at ( j );
            }
            result[ ( i * windowOffset ) + j ] = result.at( ( i * windowOffset ) + j ) + currentWindow.at ( j );
        }
    }

    //std::cerr << "result.size() = " << result.size() << std::endl;
    channel = result;

}

template<class T>
T SliceDistortTransformer<T>::distortSampleProportional ( T sample, size_t pos )
{
    double absSample, linearRange;
    absSample =  fabs ( sample );
    linearRange = cutoff - distortRange;
    if ( ! ( absSample > linearRange ) )
        return amplificationFactor * sample;
    double distortFactor, absSampleDistorted, sinFactor;
    distortFactor = distortRange / ( SHRT_MAX - linearRange );
    sinFactor = sin ( 2 * PI / 2205 * pos );
    absSampleDistorted = linearRange + distortFactor * sinFactor * ( absSample - linearRange );
    return amplificationFactor * SampleMath::sign ( sample ) * absSampleDistorted;
}

#endif*/
